#########################################################################
# Copyright (c) 2002, 2003 Grgoire Weber. All Rights Reserved.
# 
# This file is part of CMFCollection.
# 
# CMFCollection is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# CMFCollection is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with CMFCollection; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#########################################################################
"""CMFCollection Tool

Provides support to manage collections:

$Id: $
"""
__version__ = "$Revision: $"

import os

import zLOG

from Globals import InitializeClass, DTMLFile
from AccessControl import ClassSecurityInfo, getSecurityManager
from Acquisition import aq_base
from OFS.Folder import Folder
from Products.CMFCore.utils import UniqueObject, getToolByName
from Products.CMFCore.ActionProviderBase import ActionProviderBase
from Products.CMFCore.ActionInformation import ActionInformation
from Products.CMFCore.Expression import Expression
from Products.CMFCore.CMFCorePermissions import ManagePortal

import Names
import FactoryDefaults

from Slots import Slots
from Helpers import Helpers
from Types import Types

_zmidir = os.path.join(os.path.dirname(__file__), 'zmi') 

class CollectionTool(Folder, ActionProviderBase):
    __doc__ = __doc__ # copy from module

    id = Names.tool_id
    meta_type = Names.tool_meta_type

    security = ClassSecurityInfo()

    manage_options = ( 
        { 'label' : 'Overview', 'action' : 'manage_overview' },
    ) + ActionProviderBase.manage_options + Folder.manage_options

    puid_recovery = 0L
    puid_blocksize = 1000L # avoids bloating ZODB
    _properties = (
        {'id': 'puid_recovery', 'type': 'long', 'mode': 'w',
         'label': "Permanent UID recovery value after zope server restart (don't change if you don't know about portal_collections internals!)"},
        {'id': 'puid_blocksize', 'type': 'long', 'mode': 'w',
         'label': 'Permanent UID recovery block size added to puid_recovery after zope server restart (changable to values higher than 1)'},
        )

    _actions = [ActionInformation(id='add_to_collection'
                                , title='Add / Propose'
                                , action=Expression(
                text='string:${object_url}/collection_addform')
                                , condition=Expression(text='python:1')
                                , permissions=(Names.ProposeContentToCollections,
                                               Names.AddContentToCollections)
                                , category='object'
                                , visible=1
                                 )]

    #
    #   build tool structure
    #

    def __init__(self, id):
        self.id = id

        # add and configure slots
        slots = self._addObject(Slots(Names.slots_id))
        slots.setConfiguration(FactoryDefaults.default_slots)

        # add and configure helpers
        helpers = self._addObject(Helpers(Names.helpers_id))
        helpers.setConfiguration(FactoryDefaults.default_helpers)

        # add and configure types
        types = self._addObject(Types(Names.types_id))
        types.setConfiguration(FactoryDefaults.default_types)

    def _addObject(self, ob):
        ### fix me: why using _objects directly, hmmm? 
        id = ob.getId()
        setattr(self, id, ob)
        self._objects = self._objects + (
            {'id': id, 'meta_type': ob.meta_type},)
        return self._getOb(id)

    security.declarePrivate('listActions')
    def listActions(self, info=None):
        """
        Return a list of action information instances 
        provided via tool
        """
        return self._actions

    security.declareProtected(ManagePortal, 'manage_overview' )
    manage_overview = DTMLFile('explainCollectionTool', _zmidir)

    ### fix me: move to own module
    _manage_workflows = DTMLFile('manage_workflows', _zmidir)

    security.declareProtected(ManagePortal, 'manage_workflows' )
    def manage_workflows(self, REQUEST, manage_tabs_message=None):
        '''
        '''
        return self._manage_workflows(REQUEST,
                                      management_view='Workflows',
                                      manage_tabs_message=manage_tabs_message,
                                     )

    #
    #   'CMF Collection Tool' interface methods
    #

    security.declarePublic('getPuid')
    def getPuid(self, obj):
        """
        Returns the permanent unique id of 'obj'.
        If the object has not yet an puid it generates one, attaches it to
        the object and returns it.
        """

        puid = self.PUID(obj)
        if puid is None:
            # not puid: set generated puid
            puid = obj.puid = self._generatePuid()

            # reindex the object
            portal_catalog = getToolByName(self, 'portal_catalog')
            portal_catalog.reindexObject(obj)

        return puid

    security.declarePublic('getObject')
    def getObject(self, puid):
        """
        Returns object reference of object having 'puid' as puid.
        """
        portal_catalog = getToolByName(self, 'portal_catalog')
        query = self.portal_catalog(puid=puid)
        len_query = len(query)

        # return None if no object found with this puid
        if len_query == 0:
            return None

        # print a message to the log  if more than one object have
        # the same puid (should not be)
        # fix me: this should be a assert
        if len_query > 1:
            zLOG.LOG("CollectionTool.getObject:", zLOG.INFO,
                     "ASSERT: %s objects have %s as puid!!!" % (len_query, puid))

        return query[0].getObject()

    security.declarePublic('PUID')
    def PUID(self, obj):
        """
        Returns the permanent unique id of 'obj' if it has one, None else.
        """
        return getattr(obj.aq_explicit, 'puid', None)

    security.declarePrivate('delPuid')
    def delPuid(self, obj):
        """
        Delete puid information in the object.

        Deleting of puid information of an object is necessary after copy&paste.
        """
        if hasattr(aq_base(obj), 'puid'):
            del obj.puid

            # reindex the object
            portal_catalog = getToolByName(self, 'portal_catalog')
            portal_catalog.reindexObject(obj)

    security.declarePublic('getImpossiblePuid')
    def getImpossiblePuid(self):
        """
        Returns a impossible permanent unique id.
        """

        return None

    def _generatePuid(self):
        """
        Returns a permanent unique id.
        """
        puid_counter = self.temp_folder.getProperty(id='puid_counter', d=None)

        # increment counter
        if puid_counter is None:
            # counter does not exist (this it the case after a zope restart)
            puid_counter = self.puid_recovery
            self.temp_folder.manage_addProperty('puid_counter', puid_counter, 'long')
        else:
            # counter exists in temp_folder
            puid_counter += 1
            self.temp_folder.manage_changeProperties(puid_counter=puid_counter)

        # increment block counter if necessary
        if puid_counter >= self.puid_recovery:
            self.puid_recovery = puid_counter + self.puid_blocksize

        return puid_counter

    def getWorkflowTool(self):
        """
        Resturns the workflow tools object.
        """
        return getToolByName(self, 'portal_workflow')

    security.declarePrivate('getCollection')
    def getCollection(self, obj):
        """
        Returns object reference if 'obj' is a collection. If 'obj'
        is folderish and contains a collection named 'index_html'
        it returns the reference to the collection with the name
        'index_html'. Returns 'None' else.
        """
        if obj.meta_type == Names.collection_meta_type:
            return obj

        # XXX: do the right thing if 'obj' is a slot!

        # XXX: shall we check if 'obj' is folderish? We should!
        if getattr(aq_base(obj), 'index_html', None) is not None and \
           getattr(aq_base(obj.index_html), 'meta_type', None) is not None and \
           obj.index_html.meta_type == Names.collection_meta_type:
                return getattr(obj, 'index_html')

        return None

    security.declarePrivate('getCollectionBase')
    def getCollectionBase(self, obj):
        """
        Returns the reference to the folderish object
        containing a 'index_html' collection when 'obj' is
        referencing the folderish object or the collection.
        If 'obj' references a non 'index_html' collection
        'obj' is returned. Returns 'None' in any other cases.
        Remark: This is usefull to generate urls without
        trailing 'index_html'.
        """
        col_obj = self.getCollection(obj)
        
        if col_obj is not None and col_obj.getId() == 'index_html':
            return col_obj.getParentNode()
        else:
            return col_obj # is None if obj is not a collection

    security.declarePrivate('getParentCollection')
    def getParentCollection(self, obj):
        """
        Returns object reference if 'obj' parent is a collection. If
        'obj' parent is folderish and contains a collection named
        'index_html' it returns the reference to 'index_html'.
        Returns 'None' else.
        """
        if obj.getId() == 'index_html':
            obj = obj.getParentNode()
        return self.getCollection(obj.getParentNode())

    security.declarePrivate('isFolderishButNotCollection')
    def isFolderishButNotCollection(self, obj):
        """
        Returns true value if 'obj' is folderish but not a collection.
        """
        obj_base = aq_base(obj)
        if getattr(obj_base, 'isPrincipiaFolderish', None) and \
           getattr(obj_base, 'meta_type', None) != Names.collection_meta_type:
            return 1
        else:
            return None

    security.declarePublic('getSiteMap')
    def getSiteMap(self, obj=None):
        """
        Returns sitemap as tuple containing object references and tree
        indentation information.
        """
        obj = obj or getToolByName(self, 'portal_url').getPortalObject()
        tree = []
        self._buildSiteTree(obj, tree)
        return tree

    EXTENDED_SITEMAP = 0x01
    BASE_REFERENCES  = 0x10

    security.declarePublic('getSiteMap')
    def getSiteMap(self, obj=None, maxdepth=None, mode=0):
        """
        Returns sitemap as tuple containing object references and tree
        indentation information.
        """
        obj = obj or getToolByName(self, 'portal_url').getPortalObject()
        tree = []

        self._buildSiteTree(obj, tree, maxdepth=maxdepth, mode=mode)
        max_depth = reduce(lambda n,m: max(n,m), map(lambda i:i[0], tree), 0)
        return (tree, max_depth)

    def _buildSiteTree(self, obj, tree, depth=0, maxdepth=None, mode=0):
        """
        Build the site tree recursively.
        """
        bind = 0

        if maxdepth is not None and maxdepth < depth:
            return bind

        # check if obj is a collection object or a folder containing
        # a index_html collection
        col_obj = self.getCollection(obj)
        if col_obj is not None:
            fol_obj = self.getCollectionBase(col_obj)
            recursed = []

            # check if collection object is a index_html
            # collection inside a folder (col_obj != fol_obj)
            if col_obj != fol_obj:
                ### index_html collection in a folderish object
                # append collections object reference to tree list
                if mode & self.BASE_REFERENCES:
                    tree.append((depth, fol_obj, 1))
                else:
                    tree.append((depth, col_obj, 1))
                recursed.append(col_obj)

                # loop over items in the index_html collection
                for s,o in col_obj.getVisibleObjects():
                    # check if item is a collection
                    o_col = self.getCollection(o)
                    if o_col is not None:
                        # get the original collection
                        o_col = s.getOriginal(o_col) or o_col
                        # register original a recursed into
                        recursed.append(o_col)
                        o_recurse = self.getCollectionBase(o_col) or o_col
                    else:
                        o_recurse = o

                    # recurse into item
                    self._buildSiteTree(o_recurse, tree, depth+1, maxdepth, mode)

                # loop over items in the folderish object
                for o in fol_obj.objectValues():
                    # check if item is a collection and not yet recursed in above
                    o_col = self.getCollection(o)
                    if o_col is None or o_col not in recursed:
                        o_recurse = o_col or o

                        # recurse into collection
                        self._buildSiteTree(o_recurse, tree, depth+1, maxdepth, mode)

            else:
                ### collection
                # append collections object reference to tree list
                if mode & self.BASE_REFERENCES:
                    tree.append((depth, fol_obj, 1))
                else:
                    tree.append((depth, col_obj, 1))

                # loop over items in the collection
                for s,o in col_obj.getVisibleObjects():
                    # check if item is a collection
                    o_col = self.getCollection(o)
                    if o_col is not None:
                        # get the original collection
                        o_recurse = s.getOriginal(o_col) or o_col
                    else:
                        o_recurse = o

                    # recurse into original collection
                    self._buildSiteTree(o_recurse, tree, depth+1, maxdepth, mode)

            bind = 1

        # check if object is folderish
        elif mode & self.EXTENDED_SITEMAP and \
             getattr(aq_base(obj), 'isPrincipiaFolderish', None):

            # save tree state
            tree_state = len(tree)

            # append folderish object reference to tree list
            tree.append((depth, obj, 0))

            # loop over items in the folderish object
            for o in obj.objectValues():
                # recurse into folder (search for collections in subtree)
                if self._buildSiteTree(o, tree, depth+1, maxdepth, mode):
                    bind = 1

            if not bind:
                # no collections in subtree --> delete folder info
                del tree[tree_state:]

        # return bind state
        return bind

    def usedWorkflows(self):
        """
        Resturns workflows used by manual collections and workflows used by registered types
        """
        portal_workflow = getToolByName(self, 'portal_workflow')

        default_chain = portal_workflow._default_chain
        cbt = portal_workflow._chains_by_type
        type_info = portal_workflow._listTypeInfo()

        workflow_info = {}
        for ti in type_info:
            id = ti.getId()
            if cbt is not None and cbt.has_key(id):
                workflows = cbt[id]
            else:
                workflows = default_chain

            if id == 'Collection':
                collection_workflows = workflows

            for w in workflows:
                if workflow_info.has_key(w):
                    workflow_info[w].append(id)
                else:
                    workflow_info[w] = [id]

        return (collection_workflows, workflow_info)

    def getStatesOfWorkflow(self, workflow_id):
        """
        Returns a list of states defined by teh specified workflow
        """
        portal_workflow = getToolByName(self, 'portal_workflow')
        return getattr(portal_workflow, workflow_id).states.objectIds()

    def isStateInMatrixSet(self, col_wf, obj_wf, col_state, obj_state, state):
        """
        Returns state set in the matrix.
        """
        try:
            return self.matrix[col_wf][obj_wf][col_state][obj_state] == state
        except:
            return obj_state == state

    def manage_updateStateMatrix(self, REQUEST):
        """
        Updates the state matrix according to the values set in the REQUEST.
        """
        portal_workflow = getToolByName(self, 'portal_workflow')

        form = REQUEST.form
        (col_wf, obj_wf) = self.usedWorkflows()

        matrix = {}

        for cwf in col_wf:
            if not matrix.has_key(cwf):
                matrix[cwf] = {}
            matrix_cwf = matrix[cwf]
            col_states = getattr(portal_workflow, cwf).states.objectIds()

            for owf in obj_wf.keys():
                if not matrix_cwf.has_key(owf):
                    matrix_cwf[owf] = {}
                matrix_cwf_owf = matrix_cwf[owf]
                obj_states = getattr(portal_workflow, owf).states.objectIds()

                for cst in col_states:
                    if not matrix_cwf_owf.has_key(cst):
                        matrix_cwf_owf[cst] = {}
                    matrix_cwf_owf_cst = matrix_cwf_owf[cst]

                    for ost in obj_states:
                        matrix_cwf_owf_cst[ost] = form['/'.join((cwf,owf,cst,ost))]

        self.matrix = matrix

        

InitializeClass(CollectionTool)
