#########################################################################
# Copyright (c) 2002, 2003 Grgoire Weber. All Rights Reserved.
# 
# This file is part of CMFCollection.
# 
# CMFCollection is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# CMFCollection is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with CMFCollection; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#########################################################################
""" Type definitions

Definition and configuration of types beeing used by various collections.

$Id: $
"""
__version__ = "$Revision: $"

from copy import deepcopy

from Globals import InitializeClass, DTMLFile
from AccessControl import ClassSecurityInfo
from Acquisition import aq_base, aq_parent, aq_inner
from OFS.Folder import Folder

from Products.CMFCore.utils import SimpleItemWithProperties, getToolByName
from Products.CMFCore.CMFCorePermissions import ManagePortal
from Products.CMFCore.TypesTool import FactoryTypeInformation

import FactoryDefaults 
import Names

from Products.DCWorkflow.ContainerTab import ContainerTab
from CollectionType import CollectionType, factory_type_information
from Configuration import Configuration
from Helpers import Helpers
from Slots import Slots

class HelperConfiguration(SimpleItemWithProperties, Configuration):
    meta_type = 'Collection Type Helper Configuration'

    def __init__(self, id):
        self.id = id

    def getProperty(self, id, d=None):
        """Override 'PropertyManager's getProperty method to allow
           dynamically generated properties. Caution: For read-only
           properties only!"""
        if self.hasProperty(id):
            prop = getattr(self, id)
            if callable(prop):
                return prop()
            else:
                return prop
        return d

    _properties = ({'id':'title', 'type':'string', 'mode':''},
                   {'id':'allowed_templates', 'type':'multiple selection', 'mode':'w', 'select_variable':'getAllowedTemplates'},
                   {'id':'default_template', 'type':'selection', 'mode':'w', 'select_variable':'getChoosenTemplates'},
                  )

    def getAllowedTemplates(self):
        portal_collections = getToolByName(self,'portal_collections')
        helpers = getattr(portal_collections, Names.helpers_id)
        helper_templates = getattr(aq_base(helpers), self.getId())
        return helper_templates.objectIds()

    def getChoosenTemplates(self):
        return self.allowed_templates

    def title(self):
        """This is a callable property.
        """
        portal_collections = getToolByName(self,'portal_collections')
        helpers = getattr(portal_collections, Names.helpers_id)
        helper_template = getattr(aq_base(helpers), self.getId())
        return helper_template.getProperty('title', None)

    allowed_templates = []
    default_template = None

InitializeClass(HelperConfiguration)


class SlotConfiguration(ContainerTab, Configuration):
    meta_type = 'Collection Type Slot Configuration'

    def getProperty(self, id, d=None):
        """Override 'PropertyManager's getProperty method to allow
           dynamically generated properties. Caution: For read-only
           properties only!"""
        if self.hasProperty(id):
            prop = getattr(self, id)
            if callable(prop):
                return prop()
            else:
                return prop
        return d

    _properties = ({'id':'title', 'type':'string', 'mode':''},
                   {'id':'default_title_visibility', 'type':'boolean', 'mode':'w'},
                   {'id':'default_description_visibility', 'type':'boolean', 'mode':'w'},
                   {'id':'default_body_visibility', 'type':'boolean', 'mode':'w'},
                   {'id':'ordering', 'type':'selection', 'mode':'w', 'select_variable':'getOrderCriterias'},
                   {'id':'initial_position', 'type':'selection', 'mode':'w', 'select_variable':'initialPositions'},
                  )

    def title(self):
        """This is a callable property.
        """
        portal_collections = getToolByName(self,'portal_collections')
        slots = getattr(portal_collections, Names.slots_id)
        slot = getattr(aq_base(slots), self.getId())
        return slot.getProperty('title', None)

    default_title_visibility = 0
    default_description_visibility = 0
    default_body_visibility = 0

    _manual_ordering = 'manual'
    ordering = _manual_ordering

    initialPositions = ['top', 'botttom']
    initial_position = initialPositions[0]

    security = ClassSecurityInfo()
    security.declareObjectProtected(ManagePortal)

    all_meta_types = ({'name':HelperConfiguration.meta_type,
                       'action':'manage_addHelperConfigForm',
                       },)

    manage_options = Folder.manage_options

    manage_addHelperConfigForm = DTMLFile('zmi/addHelperConfigForm', globals())

    def getOrderCriterias(self):
        ### fix me: allow automatic ordering
        return (self._manual_ordering,)

    def getAvailableHelpers(self):
        """Returns a list of Helpers available to this slot
        """
        return self.objectIds()

    def getAddableHelpers(self):
        """Returns a list of all addable helpers (the one defined under
           'templates_id').
        """
        portal_collections = getToolByName(self,'portal_collections')
        helpers = getattr(portal_collections, Names.helpers_id)
        already_added = self.getAvailableHelpers()
        addable = []
        for id in helpers.objectIds():
            if not id in already_added:
                addable.append(id)
        return tuple(addable)

    def _addObject(self, id):
        obj = HelperConfiguration(id)
        self._setObject(id, obj)
        return self._getOb(id)

    def manage_addHelperConfig(self, ids=[], REQUEST=None):
        """Add a helper
        """
        for id in ids:
            self._addObject(id)
        if REQUEST is not None:
            return self.manage_main(REQUEST, 'Helper added.')

InitializeClass(SlotConfiguration)


class TypeDefinition(ContainerTab, Configuration):
    meta_type = 'Collection Type'

    _collection_properties = ({'id':'initial_slot_added',
                               'type':'selection', 'mode':'w',
                               'select_variable':'getAvailableSlots'},
                              {'id':'initial_slot_proposed',
                               'type':'selection', 'mode':'w',
                               'select_variable':'getAvailableSlots'},
                              {'id':'link_policy',
                               'type':'selection', 'mode':'w',
                               'select_variable':'getAvailableLinkPolicies'},
                              {'id':'auto_update',
                               'type':'boolean', 'mode':'w'},
                              {'id':'track_changes',
                               'type':'boolean', 'mode':'w'},
                              {'id':'verbose_status',
                               'type':'boolean', 'mode':'w'},
                              {'id':'body_prefix',
                               'type':'string', 'mode':'w'},
                             )

    initial_slot_added = None
    initial_slot_proposed = None

    def getAvailableLinkPolicies(self=None):
        """Returns a list of available link policies.
        """
        return Names.LinkPolicies

    link_policy = Names.CopyAlways

    auto_update = 0

    track_changes = 0

    verbose_status = 0

    body_prefix = 'renderBody_'

    # build _properties
    _insert_after = 'description'
    _properties = []
    for fti_p in FactoryTypeInformation._properties:
        p = deepcopy(fti_p)
        # make properties from portal_types read only
        p['mode'] = ''
        _properties.append(p)
        if p['id'] == _insert_after:
            _properties.extend(_collection_properties)
    _properties = tuple(_properties)

    def getProperty(self, id, d=None):
        """Override getProperty from 'PropertyManager' to allow
           read only remote properties. Caution: For read-only
           properties only!"""
        # try first getting own properties
        if self.hasProperty(id) and id not in ('title', 'description'):
            try:
                return getattr(self, id)
            except AttributeError:
                pass
        # else return the original values from the portal_types tool

        portal_types = getToolByName(self, 'portal_types')
        collection = getattr(aq_base(portal_types), self.getId(), None)
        if collection is not None:
            return collection.getProperty(id, d)
        return d

    security = ClassSecurityInfo()
    security.declareObjectProtected(ManagePortal)

    all_meta_types = ({'name':SlotConfiguration.meta_type,
                       'action':'manage_addTypesSlotForm',
                       },)

    manage_options = Folder.manage_options

    manage_addTypesSlotForm = DTMLFile('zmi/addSlotConfigForm', globals())

    def manage_beforeDelete(self, item, container):
        if self is item:
            try:
                portal_types = getToolByName(self, 'portal_types')
                portal_types.manage_delObjects(ids=[self.getId()])
            except (AttributeError, 'BadRequest'):
                pass

    def _addObject(self, id):
        obj = SlotConfiguration(id)
        self._setObject(id, obj)
        return self._getOb(id)

    def manage_addSlotConfig(self, ids=[], REQUEST=None):
        """Add a slot
        """
        for id in ids:
            self._addObject(id)
        if REQUEST is not None:
            return self.manage_main(REQUEST, 'Slot added.')

    def getAvailableSlots(self):
        """Returns a list of slots available to this collection type
        """
        return self.objectIds()

    def getAddableSlots(self):
        """Returns a list of all addable slots (the one defined under
           'slots_id').
        """
        portal_collections = getToolByName(self,'portal_collections')
        slots = getattr(portal_collections, Names.slots_id)
        already_added = self.getAvailableSlots()
        addable = []
        for id in slots.objectIds():
            if not id in already_added:
                addable.append(id)
        return tuple(addable)

InitializeClass(TypeDefinition)


class Types(ContainerTab, Configuration):
    meta_type = 'Collection Types'

    def __init__(self, id):
        ContainerTab.__init__(self, id)

    security = ClassSecurityInfo()
    security.declareObjectProtected(ManagePortal)

    all_meta_types = ({'name':TypeDefinition.meta_type,
                       'action':'manage_addTypeDefinitionForm',
                       },)

    manage_options = Folder.manage_options

    manage_addTypeDefinitionForm = DTMLFile('zmi/addTypeForm', globals())

    def _buildTypeInfoName(self):
        mt = CollectionType.meta_type
        return "%s: %s" % (
            filter(
                lambda fti,mt=mt:fti['meta_type']==mt,
                factory_type_information)[0]['product'],
            mt)

    def _addObject(self, id):
        # only add new type to portal_types if it doesn't exist yet
        try:
            portal_types = getToolByName(self, 'portal_types')
        except AttributeError:
            # portal_types can not be found at installation time
            # of CMFCollection (this is normal!)
            pass
        else:
            existing_obj = getattr(aq_base(portal_types), id, None)
            if existing_obj is not None:
                # raise exception if the existing type is not of type Collection
                if existing_obj.Metatype() != CollectionType.meta_type:
                    raise BadRequestException, ('The id "%s" is invalid--it is already used by another type of content type.' % id)
            else:
                # add new Factory-based Type information
                portal_types.manage_addTypeInformation(
                    FactoryTypeInformation.meta_type,
                    id=id,
                    typeinfo_name=self._buildTypeInfoName(),
                    )
                getattr(portal_types, id)._updateProperty('title', id)

        obj = TypeDefinition(id)
        self._setObject(id, obj)
        return self._getOb(id)

    def manage_addType(self, id, REQUEST=None):
        """Add a type
        """
        self._addObject(id)
        if REQUEST is not None:
            return self.manage_main(REQUEST, 'Collection Type added.')

InitializeClass(Types)
